%IKFD Indefinite Kernel Fisher Discriminant
% - with Nystroem approximation
%
%   W = IKFD(ny,ALF,BIAS)
%
% INPUT
%   K     nystroem cell ny{1} = N x m, ny{2} = (m x m)^(-1), ny{3} =
%   Landmarks; ny{4} = Labellist
%   lab   crisp label vector (continous 1 ... C)
%   ALF   Regularization constant
%   BIAS  Use bias (1) or not (0)
%
% OUTPUT
%   W     Trained Kernel Fisher discriminant
%
% DEFAULT
%   ALF  = 0.0001
%   BIAS = 1
%
% DESCRIPTION
% Finds a Fisher linear discriminant in a kernel-induced space.
% Regularization is necessary.
% Multi-class classifier is trained one-vs-all classes.
%
% SEE ALSO
% KPCA, KFD, FISHERC, MAPPINGS, DATASETS
%
% REFERENCE
% S. Mika, G. Ratsch, J. Weston, B. Scholkopf, and K.-R. Muller.
% Fisher discriminant analysis with kernels. In: Neural Networks for
% Signal Processing IX, pages 41-48, 1999.
%

% Copyright: Ela Pekalska, ela.pekalska@googlemail.com (original iKFD code)
% Faculty EWI, Delft University of Technology and
% School of Computer Science, University of Manchester
%
% Nystroem extension by Frank-Michael Schleif, fmschleif@googlemail.com
% University of Birmingham
% School of Computer Science
%  
% Example: (add argmaxmin)
% 
% load iris_dataset
% K=irisInputs'*irisInputs; % linear
% [Y,L] = max(irisTargets)
% Dat = dataset(K,L')
% W = ikfd(Dat,0.5,1);
% E=getdata(W);
% NrClasses = numel(E);
% mPred = zeros(size(K,1),NrClasses);
% for(k=1:NrClasses)
%     C=getdata(E{k});
%     for(i=1:size(K,1))
%         mPred(i,k) = C.rot'*K(i,:)'+C.offset;
%     end
% end
% vLabelPred = argmax(normcdf(mPred)');
% mean(vLabelPred(:) == L(:))

function W = ny_ikfd(ny,lab,alf,isb0)
if nargin < 4,
  isb0 = 1;
end

if nargin < 3 | isempty(alf),
  alf = 0.0001;
end

if alf <= 0,
  error('A small positive regularization ALF is necessary.');
end

n        = size(ny{1},1);
lablist  = unique(lab);
lablist  = lablist(:);
C        = numel(lablist);

% If more than two classes, train in the one-against-all strategy.
if C > 2,
  W = [];
  for i=1:C
    mlab = 2 - (lab == i);
    v  = ny_ikfd(ny,mlab,alf,isb0);
    W  = [W,setlabels(v(:,1),lablist(i,:))];
  end
  return

else    % Two classes
  Y  = 3 - 2 * lab;   % Set labels to +/-1
  I1 = find(Y==1);
  I2 = find(Y==-1);
  n1 = length(I1);
  n2 = length(I2);
  m = size(ny{2},1);

  % get row sums
  if(n1>0)
      M1 = (sum(ny{1}(I1,:)*ny{2},1) * ny{1}'/n1)'; %sum(KK(:,I1),2)/n1;
  else
      M1 = 0;
  end

  if(n2>0)
      M2 = (sum(ny{1}(I2,:)*ny{2},1) * ny{1}'/n2)'; %sum(KK(:,I2),2)/n2;
  else 
      M2 = 0;
  end
  R  = ny{3};
  
  % compute approximated squared K1
  ny_K1_sq = ny{1}*ny{2}*(ny{1}(I1,:)'*ny{1}(I1,:))*ny{2}*ny{1}(R,:)';
  % substract row sums

  if(n1>0)
      ny_N1{1} = ny_K1_sq - M1*M1(R)'*n1;
  end
  
  % compute approximated squared K1
  ny_K2_sq = ny{1}*ny{2}*(ny{1}(I2,:)'*ny{1}(I2,:))*ny{2}*ny{1}(R,:)';
  % substract row sums
  if(n2>0)
      ny_N2{1} = ny_K2_sq - M2*M2(R)'*n2;
  end
  
  ny_N{1} = 0;
  if(n1>0 && n2>0)
      ny_N{1}=ny_N1{1} + ny_N2{1};
  elseif(n1>0)
      ny_N{1}=ny_N1{1};
  elseif(n2>0)
      ny_N{1}=ny_N2{1};
  else
      error('both classes empty - should not happen')
  end
  ny_N{2} = pinv(ny_N{1}(R,:));
  
  if(0)
    % Regularization
    N = N + alf * eye(n);
  end
  
  % Optimization
%   %W = inv(N)*(M1-M2);       % The same as above up to scaling
%   W = V*((U'*(M1-M2))'*S)';  % using the pseudo inverse

  [C, A, idx] = eig_ny(ny_N);
  W = C*pinv(A)*(C'*(M1-M2));

  % Project data on the found direction
  b0 = 0;
  if isb0,
    % Find the free parameter
    b0 = -W'*(M1+M2)/2;
  end
end

dummy = dataset(zeros(n,1),ny{4}(:));
% Determine the mapping
W = affine(W,b0,dummy,lablist,n,2);
%W = cnormc(W,K);
%W = V*W;
W = setname(W,'IKFD');
return;
